;
;  * * * * * * * * *
;  * TAPE2CAS.COM  *
;  *   ver. 1.0    *
;  *   by Martos   *
;  * * * * * * * * *
;
; This utility reads an MSX tape and generates a CAS file out of it. 

BDOS	EQU	0005H
CALSLT	EQU	001CH
FCB	EQU	005CH
DMA	EQU	0080H
TAPION	EQU	00E1H
TAPIN	EQU	00E4H
TAPIOF	EQU	00E7H
SNSMAT	EQU	0141H
EXPTBL	EQU	0FCC1H

; To maximize the available memory for data buffer, the start-up sequence has
; been moved there, allowing it to be overwritten later and therefore adding to
; its size.

	CALL	DATA
	CALL	MAIN
	PUSH	AF
	LD	HL,(BAKPOS)
	LD	(FCB+33),HL	; Back to the last secure position
	LD	A,L
	OR	H
	LD	HL,(BAKPOS+2)
	LD	(FCB+35),HL
	OR	L
	OR	H		; Is that position = 0?
	PUSH	AF
	LD	HL,0		; Cut off unsecure data
	CALL	FILEWR
	LD	C,10H
	CALL	FILACC		; Close file
	POP	AF
	LD	C,13H
	CALL	Z,FILACC	; Delete file if length = 0
	POP	AF
	RET	Z
	DEC	A
	LD	HL,S_DISK
	JR	Z,STRING
	DEC	A
	LD	HL,S_MEM
	JR	Z,STRING
	LD	HL,S_BROK
STRING:	LD	A,(HL)		; Console output has to be done through BDOS
	OR	A		; function 06H. Any other function will abort
	RET	Z		; the program when CTRL+STOP is pressed.
	CALL	WRCHAR
	INC	HL
	JR	STRING

S_DISK:	DB	'Insufficient disk space for CAS file',0
S_MEM:	DB	'Not enough memory for a data block',0
S_BROK:	DB	'Tape file broken, or not such file',0
RAWFLG:	DS	1
BAKPOS:	DS	4

; MAIN reads data blocks from tape and adds them to output file. It will
; recognize standard tape files and check their consistence, unless RAWFLG is
; activated. Output values in A are the following, with Z flag accordingly set:
; 0: ended normally (CTRL+STOP)
; 1: file write error
; 2: data buffer overflowed
; 3: tape file broken or inconsistent

MAIN:	CALL	TAPEON
	CCF
	SBC	A,A
	RET	Z
	PUSH	DE
	LD	HL,S_RAW
	CALL	STRING
	POP	DE
	DI
RDRAW:	CALL	TAPEIN
	JR	C,CHKSFT
	JR	Z,DISCAR
	LD	A,(RAWFLG)
	OR	A
	JR	NZ,RDRAW
	LD	HL,DATA+16
	SBC	HL,DE
	JR	NZ,RDRAW
	LD	HL,DATA
	LD	A,(HL)
	INC	HL
	LD	B,9
HEADID:	CP	(HL)
	JR	NZ,RDRAW
	INC	HL
	DJNZ	HEADID
	CP	0EAH
	JR	Z,ASCII
	CP	0D3H
	JR	Z,BASIC
	CP	0D0H
	JP	Z,BINARY
	JR	RDRAW
CHKSFT:	LD	A,6
	LD	IX,SNSMAT
	CALL	BIOS
	AND	1
	JR	Z,DISCAR	; exit if SHIFT is pressed
	PUSH	DE
	CALL	ADDBLK
	POP	DE
	JR	NZ,DISCAR
	LD	HL,-DATA	-200H ; M80/L80 BUG WORKAROUND
	ADD	HL,DE		; Block length, to be displayed
	CALL	NUMBER
	LD	HL,S_BYTE
ADVPOS:	CALL	STRING
	LD	HL,FCB+33	; Current file position is secure
	LD	DE,BAKPOS
	LD	BC,4
	LDIR
	JR	MAIN
DISCAR:	OR	A
	PUSH	AF
	LD	HL,S_DISC
	CALL	STRING
	POP	AF
	RET

ASCII:	LD	HL,S_ASC
	CALL	TAPHED
	JR	NZ,DISCAR
ASCBLK:	CALL	TAPEON
	JR	C,DISCAR
	LD	B,0		; = 256 bytes
RDASC:	CALL	TAPEIN
	JR	C,DISCAR
	DJNZ	RDASC
	CALL	TAPBOD
	JR	NZ,DISCAR
	LD	A,1AH		; EOF character
	LD	HL,DATA
	LD	BC,256
	CPIR
	JR	NZ,ASCBLK	; No EOF, more ascii blocks
FILEOK:	LD	HL,S_OK
	JR	ADVPOS

BASIC:	LD	HL,S_BAS
	CALL	TAPHED
	JR	NZ,DISCAR
	CALL	TAPEON
	JR	C,DISCAR
RESCNT:	LD	B,10
RDBAS:	CALL	TAPEIN
	JR	C,DISCAR
	JR	Z,DISCAR
	OR	A
	JR	NZ,RESCNT
	DJNZ	RDBAS		; 10th consecutive 00 read?
SAVBOD:	CALL	TAPBOD
	JR	NZ,DISCAR
	JR	FILEOK

BINARY:	LD	HL,S_BIN
	CALL	TAPHED
	JR	NZ,DISCAR
	CALL	TAPEON
	JR	C,DISCAR
	LD	B,6
RDBIN1:	CALL	TAPEIN
	JR	C,DISCAR
	DJNZ	RDBIN1
	LD	HL,(DATA+2)	; Last address
	LD	BC,(DATA)	; First address
	SBC	HL,BC
	CALL	C,TAPERR	; A = 3 and tape off, flags unaffected
	JR	C,DISCAR
	INC	HL
	LD	B,H
	LD	C,L
RDBIN2:	CALL	TAPEIN
	JP	C,DISCAR
	JP	Z,DISCAR
	DEC	BC
	LD	A,B
	OR	C
	JR	NZ,RDBIN2
	JR	SAVBOD

NUMBER:	LD	C,0		; Output of initial zeros disabled
	LD	DE,10000
	CALL	DIGIT
	LD	DE,1000
	CALL	DIGIT
	LD	DE,100
	CALL	DIGIT
	LD	DE,10
	CALL	DIGIT
	LD	A,L
	ADD	A,'0'
	JR	WRCHAR

DIGIT:	XOR	A
SUBSTR:	SBC	HL,DE
	JR	C,REMAIN
	INC	A
	LD	C,'0'		; Enable output from now on
	JR	SUBSTR
REMAIN:	ADD	HL,DE		; A = quotient, HL = remainder of HL/DE
	ADD	A,C
	RET	Z		; All digits are zero up till now
WRCHAR:	LD	E,A
	PUSH	HL
	PUSH	BC
	LD	C,6		; Console output without checking CTRL+STOP
	CALL	BDOS
	POP	BC
	POP	HL
	RET

; ADDBLK adds to output file the data block in the DMA (LEADIN) up to DE-1. It
; returns A=0 (Z) if succeeded, or A=1 (NZ) if a file write error occured.
; TAPHED and TAPBOD comprise some common operations when saving a tape file's
; header or body. Both stop the tape and end up calling ADDBLK.

TAPHED:	CALL	TAPEOF
	CALL	STRING
	LD	HL,S_FILE
	CALL	STRING
	LD	HL,DATA+15	; Last character of file name
	LD	A,' '
	LD	B,5
CHKSPC:	CP	(HL)
	JR	NZ,WRNAME
	DEC	HL
	DJNZ	CHKSPC
WRNAME:	INC	B		; Name length discarding spaces
	LD	HL,DATA+10
NAMCHR:	LD	A,(HL)
	CALL	WRCHAR
	INC	HL
	DJNZ	NAMCHR
	LD	HL,S_QUOT
	CALL	STRING
	LD	DE,DATA+16
	JR	ADDBLK
TAPBOD:	CALL	TAPEOF
ADDBLK:	LD	HL,FCB+33
	LD	A,(HL)
	AND	0F8H
	CP	(HL)
	JR	Z,WRBLK		; Already 8-byte alligned
	ADD	A,8
	LD	(HL),A
	JR	NC,WRBLK
	LD	B,3
PROPCY:	INC	HL
	INC	(HL)		; C flag is unaffected,
	JR	NZ,WRBLK	; but Z indicates the same
	DJNZ	PROPCY
WRBLK:	LD	HL,-LEADIN	-200H ; M80/L80 BUG WORKAROUND
	ADD	HL,DE		; Block length, including CAS-format header
FILEWR:	LD	C,26H
FILACC:	LD	DE,FCB
	CALL	BDOS
	OR	A
	RET

; Standard tape read functions are accessed through TAPEON, TAPEIN and TAPEOF,
; which also perform several related operations. TAPEIN stores the byte read at
; (DE) and increases DE, unless a buffer overflow occured (noted with A=2 and
; NC,Z). TAPEON initializes DE to the buffer's start address. Both will return
; A=3 and C flag activated in case of a tape read error.

TAPEIN:	PUSH	BC
	PUSH	DE
	LD	IX,TAPIN
	CALL	BIOS
	POP	DE
	POP	BC
	JR	C,TAPERR
	LD	HL,(BDOS+1)	; End of TPA
	SBC	HL,DE
	JR	Z,MEMOUT
	LD	(DE),A
	INC	DE
	RET
MEMOUT:	LD	A,2
	JR	TAPEOF
TAPEON:	LD	IX,TAPION
	CALL	BIOS
	LD	DE,DATA
	RET	NC
TAPERR:	LD	A,3
TAPEOF:	LD	IX,TAPIOF
BIOS:	LD	IY,(EXPTBL-1)
	JP	CALSLT

S_RAW:	DB	'Raw data: ',0
S_BYTE:	DB	' bytes read',13,10,0
S_ASC:	DB	13,'Ascii',0
S_BAS:	DB	13,'Basic',0
S_BIN:	DB	13,'Binary',0
S_FILE:	DB	' file "',0
S_QUOT:	DB	'": ',0
S_OK:	DB	'OK',13,10,0
S_DISC:	DB	'discarded',13,10,0

LEADIN:	DB	1FH,0A6H,0DEH,0BAH,0CCH,13H,7DH,74H
DATA:

; Start-up sequence, put here for allowing its space to be used later for data.

	LD	A,(FCB+1)
	CP	' '
	LD	HL,S_INST
	JR	Z,EXIT		; Invalid arguments => show instructions
	LD	HL,FCB+9
	LD	A,(HL)
	CP	' '
	JR	NZ,SRCHSW
	LD	(HL),'C'	; No extension => set it to .CAS
	INC	HL
	LD	(HL),'A'
	INC	HL
	LD	(HL),'S'
SRCHSW:	LD	DE,0016H	; Default values in case of no switches
	LD	HL,DMA
	LD	C,(HL)
	LD	B,0
	INC	HL
SWCHK1:	LD	A,'/'
	CPIR
	JP	PO,INITSW
	LD	A,(HL)
	AND	0DFH		; To capital letters
	CP	'R'
	JR	NZ,SWCHK2
	LD	D,1
SWCHK2:	CP	'A'
	JR	NZ,SWCHK1
	LD	E,0FH
	JR	SWCHK1
INITSW:	LD	A,D		; 0, or set to 1 if [/R]aw mode
	LD	(RAWFLG),A
	LD	C,E		; Create (16H), or just open (0FH) if [/A]ppend
	CALL	FILACC
	LD	HL,S_OPEN
	JR	NZ,EXIT
	LD	HL,1
	LD	(FCB+14),HL	; Register length = 1
	LD	HL,(FCB+16)
	LD	(FCB+33),HL	; Set register pointer to the end of file
	LD	(BAKPOS),HL	; Previous contents will never be cut off
	LD	HL,(FCB+18)
	LD	(FCB+35),HL
	LD	(BAKPOS+2),HL
	LD	DE,LEADIN
	LD	C,1AH
	JP	BDOS		; Set the DMA to LEADIN + following DATA
EXIT:	POP	AF
	JP	STRING

S_INST:	DB	10
	DB	'  TAPE2CAS <output file> [/R][/A]',13,10
	DB	10
	DB	"[/R]aw mode, don't parse tape files",13,10
	DB	'[/A]ppend to an existing CAS file',13,10
	DB	10
	DB	'CTRL+STOP: force the end of current',13,10
	DB	'  data. In case of a raw block, it',13,10
	DB	'  is saved and the program goes on.',13,10
	DB	'SHIFT+CTRL+STOP: terminate and exit',13,10
	DB	'  in any case.',13,10,0
S_OPEN:	DB	"Can't open output file",0

	END
